/*
  GPSJinni - show raw data from the GPS subsystem.
  Copyright (C) 2009  Tim Teulings

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <limits>

#include "config.h"

#include <signal.h>

#include <glib.h>

#include <cerrno>
#include <cmath>

#if defined (HAVE_LIB_GPS)
  #include <gps.h>
#endif

#if defined(HAVE_LIB_LOCATION)
extern "C" {
  #include <location/location-gps-device.h>
  #include <location/location-gpsd-control.h>
}
#endif

#include <netinet/in.h>

#include <fstream>
#include <iomanip>
#include <sstream>

#include <Lum/Def/Menu.h>
#include <Lum/Def/MultiView.h>

#include <Lum/OS/Probe.h>
#include <Lum/OS/Thread.h>

#include <Lum/Base/DateTime.h>
#include <Lum/Base/L10N.h>
#include <Lum/Base/Path.h>
#include <Lum/Base/String.h>

#include <Lum/Dlg/About.h>
#include <Lum/Dlg/File.h>
#include <Lum/Dlg/Msg.h>
#include <Lum/Dlg/Value.h>

#include <Lum/Model/Action.h>
#include <Lum/Model/Boolean.h>
#include <Lum/Model/DataStream.h>
#include <Lum/Model/Header.h>
#include <Lum/Model/String.h>

#include <Lum/OS/Main.h>

#include <Lum/Button.h>
#include <Lum/Dialog.h>
#include <Lum/Label.h>
#include <Lum/LED.h>
#include <Lum/Panel.h>
#include <Lum/PercentBar.h>
#include <Lum/Tab.h>
#include <Lum/Table.h>
#include <Lum/Text.h>
#include <Lum/TextValue.h>
#include <Lum/View.h>

#include "Compass.h"
#include "TimeLine.h"

#include <iostream>

// maximum age of data in seconds until it is invalid
#if defined(HAVE_LIB_GPS)
// Using normal libgps we don't trust situation if GPS daemon is not
// responding in time
static long obsoleteDataAge = 30;
#else
// Using liblocation on the other hand we trust it to send diffs
// every time something relvant changes.
// Data does not age!
static long obsoleteDataAge = std::numeric_limits<long>::max();
#endif

static long reconnectTime   = 10; // Period of time after which we check for disconnection and
                                  // trigger reconnection
static long refreshTime     = 10; // Period of time after which we update the stream views

class GPSData
{
public:
  struct Satellite
  {
    int  id;
    bool used;
    int  snr;
  };

private:
  bool                   connected;
  Lum::Base::SystemTime  connectedLastData;

  bool                   online;
  Lum::Base::SystemTime  onlineLastData;

  bool                   fix;
  Lum::Base::SystemTime  fixLastData;

  double                 latitude;
  Lum::Base::SystemTime  latitudeLastData;

  double                 longitude;
  Lum::Base::SystemTime  longitudeLastData;

  double                 altitude;
  Lum::Base::SystemTime  altitudeLastData;

  double                 speed;
  Lum::Base::SystemTime  speedLastData;

  double                 track;
  Lum::Base::SystemTime  trackLastData;

  double                 climb;
  Lum::Base::SystemTime  climbLastData;

  Lum::Base::SystemTime  time;
  Lum::Base::SystemTime  timeLastData;

  std::vector<Satellite> satellites;
  Lum::Base::SystemTime  satellitesLastData;

public:
  GPSData()
   : connectedLastData(0),
     onlineLastData(0),
     fixLastData(0),
     latitudeLastData(0),
     longitudeLastData(0),
     altitudeLastData(0),
     speedLastData(0),
     trackLastData(0),
     climbLastData(0),
     timeLastData(0),
     satellitesLastData(0)
  {
    // no code
  }

  //
  // Connected
  //
  void SetConnected(bool connected)
  {
    connectedLastData.SetToNow();
    this->connected=connected;
  }

  bool HasConnected() const
  {
    Lum::Base::SystemTime now;

    return now-connectedLastData<obsoleteDataAge;
  }

  bool GetConnected() const
  {
    assert(HasConnected());

    return connected;
  }

  //
  // Online
  //
  void SetOnline(bool online)
  {
    onlineLastData.SetToNow();
    this->online=online;
  }

  bool HasOnline() const
  {
    Lum::Base::SystemTime now;

    return now-onlineLastData<obsoleteDataAge;
  }

  bool GetOnline() const
  {
    assert(HasOnline());

    return online;
  }

  //
  // Fix
  //
  void SetFix(bool fix)
  {
    fixLastData.SetToNow();
    this->fix=fix;
  }

  bool HasFix() const
  {
    Lum::Base::SystemTime now;

    return now-fixLastData<obsoleteDataAge;
  }

  bool GetFix() const
  {
    assert(HasFix());

    return fix;
  }

  //
  // Latitude
  //
  void SetLatitude(double latitude)
  {
    latitudeLastData.SetToNow();
    this->latitude=latitude;
  }

  bool HasLatitude() const
  {
    Lum::Base::SystemTime now;

    return now-latitudeLastData<obsoleteDataAge;
  }

  double GetLatitude() const
  {
    assert(HasLatitude());

    return latitude;
  }

  //
  // Longitude
  //
  void SetLongitude(double longitude)
  {
    longitudeLastData.SetToNow();
    this->longitude=longitude;
  }

  bool HasLongitude() const
  {
    Lum::Base::SystemTime now;

    return now-longitudeLastData<obsoleteDataAge;
  }

  double GetLongitude() const
  {
    assert(HasLongitude());

    return longitude;
  }

  //
  // Altitude
  //
  void SetAltitude(double altitude)
  {
    altitudeLastData.SetToNow();
    this->altitude=altitude;
  }

  bool HasAltitude() const
  {
    Lum::Base::SystemTime now;

    return now-altitudeLastData<obsoleteDataAge;
  }

  double GetAltitude() const
  {
    assert(HasAltitude());

    return altitude;
  }

  //
  // Speed
  //
  void SetSpeed(double speed)
  {
    speedLastData.SetToNow();
    this->speed=speed;
  }

  bool HasSpeed() const
  {
    Lum::Base::SystemTime now;

    return now-speedLastData<obsoleteDataAge;
  }

  double GetSpeed() const
  {
    assert(HasSpeed());

    return speed;
  }

  //
  // Track
  //
  void SetTrack(double track)
  {
    trackLastData.SetToNow();
    this->track=track;
  }

  bool HasTrack() const
  {
    Lum::Base::SystemTime now;

    return now-trackLastData<obsoleteDataAge;
  }

  double GetTrack() const
  {
    assert(HasTrack());

    return track;
  }

  //
  // Climb
  //
  void SetClimb(double climb)
  {
    climbLastData.SetToNow();
    this->climb=climb;
  }

  bool HasClimb() const
  {
    Lum::Base::SystemTime now;

    return now-climbLastData<obsoleteDataAge;
  }

  double GetClimb() const
  {
    assert(HasClimb());

    return climb;
  }

  //
  // Time
  //
  void SetTime(Lum::Base::SystemTime& time)
  {
    timeLastData.SetToNow();
    this->time=time;
  }

  bool HasTime() const
  {
    Lum::Base::SystemTime now;

    return now-timeLastData<obsoleteDataAge;
  }

  Lum::Base::SystemTime GetTime() const
  {
    assert(HasTime());

    return time;
  }

  //
  // Satellites
  //
  void SetSatellites(const std::vector<Satellite>& satellites)
  {
    satellitesLastData.SetToNow();
    this->satellites=satellites;
  }

  bool HasSatellites() const
  {
    Lum::Base::SystemTime now;

    return now-satellitesLastData<obsoleteDataAge;
  }

  const std::vector<Satellite>& GetSatellites() const
  {
    assert(HasSatellites());

    return satellites;
  }
};

typedef Lum::Model::StdTable<GPSData::Satellite> SatellitesModel;
typedef Lum::Base::Reference<SatellitesModel>    SatellitesModelRef;

double GetRelativeQualityFromDbm(double quality)
{
  if (quality<0) {
    return 0;
  }
  if (quality>=0 && quality<40) {
    return 5*quality/2;
  }
  else {
    return 100;
  }
}

class SatelliteObject : public Lum::Object
{
private:
  Lum::OS::FontRef   font;
  Lum::OS::ColorRef  green;
  Lum::OS::ColorRef  grey;
  GPSData::Satellite satellite;

public:
  SatelliteObject()
  : font(Lum::OS::display->GetFont()),
    green(0,1,0,Lum::OS::display->GetColor(Lum::OS::Display::tableTextColor)),
    grey(0.7,0.7,0.7,Lum::OS::display->GetColor(Lum::OS::Display::tableTextColor))
  {
    // no code
  }

  void SetSatellite(const GPSData::Satellite& satellite)
  {
    this->satellite=satellite;
  }

  void CalcSize()
  {
    minWidth=Lum::OS::display->GetSpaceHorizontal(Lum::OS::Display::spaceObjectBorder)+
             10*font->StringWidth(L"m");
    minHeight=Lum::OS::display->GetSpaceVertical(Lum::OS::Display::spaceObjectBorder)+
              font->height+
              Lum::OS::display->GetSpaceVertical(Lum::OS::Display::spaceObjectBorder);

    width=minWidth;
    height=minHeight;

    Object::CalcSize();
  }

  void Draw(int x, int y, size_t w, size_t h)
  {
    Object::Draw(x,y,w,h);

    if (!OIntersect(x,y,w,h)) {
      return;
    }

    /* --- */

    Lum::OS::DrawInfo *draw=GetDrawInfo();

    //
    // Signal strength bar
    //

    double percent=GetRelativeQualityFromDbm(satellite.snr)/100.0;

    if (satellite.used) {
      draw->PushForeground(green);
    }
    else {
      draw->PushForeground(grey);
    }

    draw->FillRectangle(this->x,
                        this->y+
                        Lum::OS::display->GetSpaceVertical(Lum::OS::Display::spaceObjectBorder),
                        (size_t)(width*percent),height-2);


    draw->PopForeground();

    //
    // Name of Satellite
    //

    draw->PushFont(font);
    draw->PushForeground(parent->GetTextColor(draw));
    draw->DrawString(this->x,
                     this->y+font->ascent,
                     Lum::Base::NumberToWString(satellite.id));
    draw->PopForeground();
    draw->PopFont();

    //
    //
    // Signal strength label

    Lum::OS::FontRef    font=Lum::OS::display->GetFont();
    std::wstring        label=Lum::Base::NumberToWString(lround(percent*100))+L"%";
    Lum::OS::FontExtent extent;

    font->StringExtent(label,extent);

    draw->PushForeground(Lum::OS::Display::textColor);
    draw->PushFont(font,0);
    draw->DrawString(this->x+(width-extent.width+extent.left+extent.right)/2,
                     this->y+
                     Lum::OS::display->GetSpaceVertical(Lum::OS::Display::spaceObjectBorder)+
                     font->ascent,
                     label);
    draw->PopFont();
    draw->PopForeground();
  }

  static SatelliteObject* Create(bool horizFlex, bool vertFlex)
  {
    SatelliteObject *object=new SatelliteObject();

    object->SetFlex(horizFlex,vertFlex);

    return object;
  }
};

class SatellitesDataProvider : public SatellitesModel::DataProvider
{
private:
  SatelliteObject *object;

public:
  SatellitesDataProvider()
  : object(SatelliteObject::Create(true,false))
  {
    // no code
  }

  ~SatellitesDataProvider()
  {
    delete object;
  }

  Lum::Object* GetObject(const SatellitesModel::Iterator& iter, size_t column) const
  {
    object->SetSatellite(*iter);
    return object;
  }
};

class MyWindow;
static MyWindow*         window=NULL;
static Lum::Def::AppInfo info;
static unsigned long     trackingFileId='G'*256*256*256+'J'*256*256+'T'*256+'\0';
static unsigned long     trackingFileVersion=1;

#if defined(HAVE_LIB_GPS)
static void cb( struct gps_data_t * data,
               char * buf,
               size_t len,
               int level );
#elif defined(HAVE_LIB_LOCATION)
static void gps_location_changed(LocationGPSDevice *device, gpointer userdata);
static void gps_location_connected(LocationGPSDevice *new_device, gpointer userdata);
static void gps_location_disconnected(LocationGPSDevice *new_device, gpointer userdata);
static void gps_location_started(LocationGPSDControl *control, gpointer userdata);
static void gps_location_stopped(LocationGPSDControl *control, gpointer userdata);
static void gps_location_error(LocationGPSDControl *control, gpointer userdata);
#endif

std::wstring DoubleToString(double value)
{
  if (!isnan(value)) {
    std::stringstream buffer;

    buffer.imbue(std::locale(""));
    buffer << std::fixed;
    buffer << std::showpoint;
    buffer.precision(2);

    buffer << value;

    return Lum::Base::StringToWString(buffer.str());
  }
  else {
    return L"";
  }
}

class Track
{
private:
  std::wstring          name;
  Lum::Base::SystemTime time;

public:
  Track()
  {
    // no code
  }

  Track(const std::wstring& name, const Lum::Base::SystemTime& time)
  : name(name),
    time(time)
  {
    // no code
  }

  std::wstring GetName() const
  {
    return name;
  }

  std::wstring GetTimeString() const
  {
    return time.GetLocalLocaleDate()+L" "+time.GetLocalLocaleTime();
  }
};

class TrackObject : public Lum::Object
{
private:
  Lum::OS::FontRef font;
  Track            track;

public:
  TrackObject()
  : font(Lum::OS::display->GetFont())
  {
    // no code
  }

  void SetTrack(const Track& track)
  {
    this->track=track;
  }

  void CalcSize()
  {
    minWidth=20*font->StringWidth(L"m")+
             Lum::OS::display->GetSpaceHorizontal(Lum::OS::Display::spaceObjectBorder)+
             5*font->StringWidth(L"m");
    minHeight=font->height;

    width=minWidth;
    height=minHeight;

    Object::CalcSize();
  }

  void Draw(int x, int y, size_t w, size_t h)
  {
    Object::Draw(x,y,w,h);

    if (!OIntersect(x,y,w,h)) {
      return;
    }

    /* --- */

    Lum::OS::DrawInfo *draw=GetDrawInfo();

    draw->PushFont(font);
    draw->PushForeground(parent->GetTextColor(draw));

    //
    // Name of Track
    //

    draw->DrawString(this->x,
                     this->y+font->ascent,
                     track.GetName());

    //
    //
    // Time of track

    draw->DrawString(this->x+width-font->StringWidth(track.GetTimeString()),
                     this->y+font->ascent,
                     track.GetTimeString());

    draw->PopFont();
    draw->PopForeground();
  }

  static TrackObject* Create(bool horizFlex, bool vertFlex)
  {
    TrackObject *object=new TrackObject();

    object->SetFlex(horizFlex,vertFlex);

    return object;
  }
};

typedef Lum::Model::StdTable<Track> TracksModel;
typedef Lum::Base::Reference<TracksModel> TracksModelRef;

class TracksDataProvider : public TracksModel::DataProvider
{
private:
  TrackObject *object;

public:
  TracksDataProvider()
   : object(TrackObject::Create(true,false))
  {
    // no code
  }

  ~TracksDataProvider()
  {
    delete object;
  }

  Lum::Object* GetObject(const TracksModel::Iterator& iter, size_t column) const
  {
    object->SetTrack(*iter);
    return object;
  }
};

class TrackComparator : public TracksModel::Comparator
{
public:
  bool operator()(const Track& a, const Track& b, size_t column, bool down) const
  {
    assert(column==1);

    if (down) {
      return a.GetName()<b.GetName();
    }
    else {
      return a.GetName()>b.GetName();
    }
  }
};

class MyWindow : public Lum::Dialog
{
private:
  Lum::Model::BooleanRef   connected;
  Lum::Model::BooleanRef   online;
  Lum::Model::BooleanRef   fix;
  Lum::Model::StringRef    time;
  Lum::Model::StringRef    latitude;
  Lum::Model::StringRef    longitude;
  Lum::Model::StringRef    altitude;
  Lum::Model::StringRef    speed;
  Lum::Model::StringRef    track;
  Lum::Model::DoubleRef    direction;
  Lum::Model::StringRef    climb;
  SatellitesModelRef       satellites;
  Lum::Model::DoubleDataStreamRef speedStream;
  Lum::Model::DoubleDataStreamRef altitudeStream;
  TracksModelRef           tracks;
  Lum::Model::SingleLineSelectionRef trackSelection;

  Lum::Model::ActionRef    startTrackingAction;
  Lum::Model::ActionRef    stopTrackingAction;
  Lum::Model::ActionRef    deleteTrackAction;
  Lum::Model::ActionRef    exportTrackAction;
  Lum::Model::ActionRef    aboutAction;

  Lum::Model::ActionRef    dataChangedAction;
  Lum::Model::ActionRef    refreshTimer;
  Lum::Model::ActionRef    reconnectCheckTimer;

  Lum::Model::BooleanRef   tracking;

  std::ofstream            trackFile;

#if defined(HAVE_LIB_LOCATION)
  LocationGPSDevice        *device;
  LocationGPSDControl      *control;
#endif
#if defined(HAVE_LIB_GPS)
  struct gps_data_t        *gpsData;
  pthread_t                gpsThread;
#elif defined(HAVE_LIB_LOCATION)
    guint                  idd_changed;
    guint                  idd_con;
    guint                  idd_disc;
	
    guint                  idc_run;
    guint                  idc_stop;
    guint                  idc_error;
#endif

public:
  Lum::OS::RWMutex         dataMutex;
  GPSData                  data;

private:
  Lum::TextValue* CreateTextValue(Lum::Model::StringRef& model, size_t size)
  {
    Lum::TextValue *textValue;

    textValue=Lum::TextValue::Create(model,Lum::TextValue::left,true,false);
    textValue->SetWidth(Lum::Base::Size::stdCharWidth,size);

    return textValue;
  }

  void InitializeGPS()
  {
      std::cout << "Initializing GPS..." << std::endl;
#if defined(HAVE_LIB_LOCATION)
    std::cout << "Initialize LocationGPSDcontrol..." << std::endl;
    control=location_gpsd_control_get_default();

    if (control!=NULL) {
#if defined(LOCATION_METHOD_USER_SELECTED)
      std::cout << "Initializing LocationGPSDevice..." << std::endl;
      device=(LocationGPSDevice*)g_object_new(LOCATION_TYPE_GPS_DEVICE, NULL);

      if (device!=NULL) {
        std::cout << "Connecting callbacks..." << std::endl;

        idd_changed=g_signal_connect(device,"changed",G_CALLBACK(gps_location_changed),NULL);
        idd_con=g_signal_connect(device,"connected",G_CALLBACK(gps_location_connected),NULL);
        idd_disc=g_signal_connect(device,"disconnected",G_CALLBACK(gps_location_disconnected),NULL);

        idc_run=g_signal_connect(control,"gpsd_running",G_CALLBACK(gps_location_started),NULL);
        idc_stop=g_signal_connect(control,"gpsd_stopped",G_CALLBACK(gps_location_stopped),NULL);
        idc_error=g_signal_connect(control,"error",G_CALLBACK(gps_location_error),NULL);
      }
      else {
        std::cerr << "GPS NOT initialized!" << std::endl;
      }

      g_object_set(G_OBJECT(control),
                   "preferred-method",
                   LOCATION_METHOD_USER_SELECTED,
                   NULL);
      g_object_set(G_OBJECT(control),
                   "preferred-interval",
                   LOCATION_INTERVAL_5S,
                   NULL);
#endif
      std::cout << "Starting GPS daemon..." << std::endl;
      location_gpsd_control_start(control);
    }
    else {
      std::cerr << "Cannot start GPS daemon!" << std::endl;
    }
#endif

#if defined(HAVE_LIB_GPS)
    std::cout << "Initializing GPS daemon..." << std::endl;
    // This fixed a bug in libgps, using locale aware functions to parse non-locale-aware float values
    setlocale(LC_NUMERIC,"C");
    gpsData=::gps_open("127.0.0.1","2947");

    if (gpsData!=NULL) {
      gps_query(gpsData,"j=0w+x\n");
      gps_set_callback(gpsData,cb,&gpsThread);
      std::cout << "GPS initialized!" << std::endl;
    }
    else {
      std::cerr << "GPS NOT initialized: " << errno << std::endl;
    }
#elif !defined(HAVE_LIB_LOCATION)
    std::cerr << "No GPS initialization possible!" << std::endl;
#endif
  }

  void DeinitializeGPS()
  {
    std::cout << "Deinitialize GPS..." << std::endl;

#if defined(HAVE_LIB_GPS)
    if (gpsData!=NULL) {
      ::gps_close(gpsData);
      gpsData=NULL;
    }
#elif defined(HAVE_LIB_LOCATION)
    if (device!=NULL) {
      g_signal_handler_disconnect(device,idd_changed);
      g_signal_handler_disconnect(device,idd_con);
      g_signal_handler_disconnect(device,idd_disc);
  	
      g_signal_handler_disconnect(control,idc_run);
      g_signal_handler_disconnect(control,idc_stop);
      g_signal_handler_disconnect(control,idc_error);

      std::cout << "Freeing LocationGPSDevice..." << std::endl;
      g_object_unref(device);	
    }
#endif

#if defined(HAVE_LIB_LOCATION)
    if (control!=NULL) {
      std::cout << "Freeing LocationGPSDcontrol..." << std::endl;

      location_gpsd_control_stop(control);

      /*Device control cleanup */
      g_object_unref(control);
      control=NULL;
    }
#endif
    std::cout << "GPS deinitialized!" << std::endl;
  }

  void ReconnectIfDisconnected()
  {
    if (data.HasConnected() && data.GetConnected()) {
      return;
    }

#if defined(HAVE_LIB_GPS)
    if (gpsData!=NULL) {
      ::gps_close(gpsData);
      gpsData=NULL;
    }

    gpsData=::gps_open("127.0.0.1","2947");

    if (gpsData!=NULL) {
      gps_query(gpsData,"j=0w+x\n");
      gps_set_callback(gpsData,cb,&gpsThread);
      data.SetConnected(true);
      std::cout << "GPS initialized!" << std::endl;
    }
    else {
      data.SetConnected(false);
      std::cout << "GPS NOT initialized: " << errno << std::endl;
    }
#endif
  }

public:
  MyWindow()
  : connected(new Lum::Model::Boolean(false)),
    online(new Lum::Model::Boolean(false)),
    fix(new Lum::Model::Boolean(false)),
    time(new Lum::Model::String(L"")),
    latitude(new Lum::Model::String(L"")),
    longitude(new Lum::Model::String(L"")),
    altitude(new Lum::Model::String(L"")),
    speed(new Lum::Model::String(L"")),
    track(new Lum::Model::String(L"")),
    direction(new Lum::Model::Double()),
    climb(new Lum::Model::String(L"")),
    satellites(new SatellitesModel(new SatellitesDataProvider())),
    speedStream(new Lum::Model::DoubleDataStream()),
    altitudeStream(new Lum::Model::DoubleDataStream()),
    tracks(new TracksModel(new TracksDataProvider(),new TrackComparator())),
    trackSelection(new Lum::Model::SingleLineSelection()),
    startTrackingAction(new Lum::Model::Action()),
    stopTrackingAction(new Lum::Model::Action()),
    deleteTrackAction(new Lum::Model::Action()),
    exportTrackAction(new Lum::Model::Action()),
    aboutAction(new Lum::Model::Action()),
    dataChangedAction(new Lum::Model::Action()),
    refreshTimer(new Lum::Model::Action()),
    reconnectCheckTimer(new Lum::Model::Action()),
    tracking(new Lum::Model::Boolean(false))
#if defined(HAVE_LIB_LOCATION)
    ,device(NULL)
    ,control(NULL)
#endif
#if defined(HAVE_LIB_GPS)
    ,gpsData(NULL)
#elif defined(HAVE_LIB_LOCATION)
    ,idd_changed(0)
    ,idd_con(0)
    ,idd_disc(0)
    ,idc_run(0)
    ,idc_stop(0)
    ,idc_error(0)
#endif
  {
    Observe(GetOpenedAction());
    Observe(GetClosedAction());
    Observe(startTrackingAction);
    Observe(stopTrackingAction);
    Observe(deleteTrackAction);
    Observe(exportTrackAction);
    Observe(aboutAction);

    Observe(dataChangedAction);
    Observe(refreshTimer);
    Observe(reconnectCheckTimer);
    Observe(trackSelection);

    ::window=this;

    startTrackingAction->Enable();
    stopTrackingAction->Disable();

    satellites->SetEmptyText(L"(no satellites found)");
    tracks->SetEmptyText(L"(no tracks found)");

    // For testsing purposes
    /*
    for (int i=0; i<20; i++) {
      GPSData::Satellite s;

      s.id=i;
      s.used=i%2!=0;
      s.snr=20;

      satellites->Append(s);
    }*/
  }

  ~MyWindow()
  {
    if (tracking->Get()) {
      StopTracking();
    }
  }

  void PreInit()
  {
    Lum::Model::HeaderRef headerModel;
    Lum::Label            *label;
    Lum::Panel            *hPanel;
    Lum::Table            *table;
    Compass               *compass;

    Lum::Def::MultiView multiView(Lum::Def::Desc(L"GPSJinni"));

    hPanel=Lum::HPanel::Create(true,true);

    label=Lum::Label::Create(true,true);

    label->AddLabel(L"Conn./Online/Fix:",
                    Lum::HPanel::Create()
                    ->Add(new Lum::LED(connected))
                    ->AddSpace()
                    ->Add(new Lum::LED(online))
                    ->AddSpace()
                    ->Add(new Lum::LED(fix)));
    label->AddLabel(L"Time:",CreateTextValue(time,24)); // TODO: Measure by example
    label->AddLabel(L"Latitude:",CreateTextValue(latitude,10));
    label->AddLabel(L"Longitude:",CreateTextValue(longitude,10));
    label->AddLabel(L"Altitude:",CreateTextValue(altitude,10));
    label->AddLabel(L"Speed:",CreateTextValue(speed,10));
    label->AddLabel(L"Direction:",CreateTextValue(track,10));
    label->AddLabel(L"Climb:",CreateTextValue(climb,10));
    hPanel->Add(label);

    multiView.AddView(Lum::Def::MultiView::View(0,
                                                Lum::Def::Desc(L"General"),
                                                hPanel));

    headerModel=new Lum::Model::HeaderImpl();
    headerModel->AddColumn(L"Satellite",Lum::Base::Size::stdCharWidth,25,true);

    table=new Lum::Table();
    table->SetFlex(true,true);
    table->SetShowHeader(false);
    table->SetHeaderModel(headerModel);
    table->SetModel(satellites);
    table->GetTableView()->SetAutoFitColumns(true);

    multiView.AddView(Lum::Def::MultiView::View(1,
                                                Lum::Def::Desc(L"Satellites"),
                                                table));

    compass=new Compass();
    compass->SetFlex(true,true);
    compass->SetModel(direction);

    multiView.AddView(Lum::Def::MultiView::View(2,
                                                Lum::Def::Desc(L"Compass"),
                                                compass));

    multiView.AddView(Lum::Def::MultiView::View(3,
                                                Lum::Def::Desc(L"Diagrams"),
                                                Lum::VPanel::Create(true,true)
                                                ->Add(Lum::Text::Create(L"Speed (m/s)"))
                                                ->AddSpace()
                                                ->Add(Lum::View::Create(TimeLine::Create(speedStream,true,true),true,true))
                                                ->AddSpace()
                                                ->Add(Lum::Text::Create(L"Altitude (m)"))
                                                ->AddSpace()
                                                ->Add(Lum::View::Create(TimeLine::Create(altitudeStream,true,true),true,true))));

    headerModel=new Lum::Model::HeaderImpl();
    headerModel->AddColumn(L"Track",Lum::Base::Size::stdCharWidth,40,true);

    table=new Lum::Table();
    table->SetFlex(true,true);
    table->SetShowHeader(false);
    table->SetHeaderModel(headerModel);
    table->SetModel(tracks);
    table->SetSelection(trackSelection);
    table->GetTableView()->SetAutoFitColumns(true);

    multiView.AddView(Lum::Def::MultiView::View(4,
                                                Lum::Def::Desc(L"Tracking"),
                                                Lum::VPanel::Create(true,true)
                                                ->Add(Lum::HPanel::Create(true,false)
                                                      ->Add(Lum::Button::Create(L"Record",
                                                                                Lum::OS::Theme::imageMediaRecord,
                                                                                startTrackingAction,
                                                                                Lum::Button::typeIcon))
                                                      ->Add(Lum::Button::Create(L"Stop",
                                                                                Lum::OS::Theme::imageMediaStop,
                                                                                stopTrackingAction,
                                                                                Lum::Button::typeIcon))
                                                      ->AddSpace(true)
                                                       ->Add(new Lum::LED(tracking))
                                                       ->AddSpace(true)
                                                      ->Add(Lum::Button::Create(L"_Delete",
                                                                                Lum::OS::Theme::imageRemove,
                                                                                deleteTrackAction,
                                                                                Lum::Button::typeIcon))
                                                      ->Add(Lum::Button::Create(L"_Export",
                                                                                Lum::OS::Theme::imageSave,
                                                                                exportTrackAction,
                                                                                Lum::Button::typeIcon)))
                                                ->AddSpace()
                                                ->Add(table)));

    Lum::Def::Menu *preMenu=Lum::Def::Menu::Create();

    preMenu
      ->GroupProject()
        ->ActionQuit(GetClosedAction())
      ->End();

    Lum::Def::Menu *postMenu=Lum::Def::Menu::Create();

    postMenu
      ->GroupHelp()
        //->ActionHelp()
        ->ActionAbout(aboutAction)
      ->End();

    multiView.SetMenu(preMenu,postMenu);

    Lum::OS::display->GetBehaviour()->ApplyMultiViewDlg(this,multiView);

    Dialog::PreInit();
  }

  void GetGPSDataCopy(GPSData& data)
  {
    Lum::OS::ReadGuard<Lum::OS::RWMutex> guard(dataMutex);

    data=this->data;
  }

  void StartTracking()
  {
    Lum::Model::StringRef trackName=new Lum::Model::String(L"");
    Lum::Base::Path       fileName;

    if (!Lum::Dlg::TextInput::GetText(GetWindow(),
                                      L"Enter track name!",
                                      L"Please enter a name for the to be recorded track.",
                                      trackName) || trackName->Empty()) {
      return;
    }

    fileName.SetNativeDir(Lum::Base::Path::GetHome());
    fileName.AppendDir(L".gpsjinni");

    Lum::Base::Path::CreateDir(fileName.GetPath());

    fileName.SetBaseName(trackName->Get()+L".gjt");

    if (Lum::Base::Path::Exists(fileName.GetPath())) {
      Lum::Dlg::Msg::ShowOk(GetWindow(),
                            L"Error while opening file!",
                            std::wstring(L"An error occured while opening file:\n")+
                            L"'"+fileName.GetPath()+L"'\n"+
                            L"Error:\n"+
                            L"'File already exists'");
      return;
    }

    trackFile.open(Lum::Base::WStringToString(fileName.GetPath()).c_str(),
                   std::ios::out|std::ios::trunc|std::ios::binary);

    if (!trackFile) {
      Lum::Base::Status status;

      status.SetToCurrentErrno();
      Lum::Dlg::Msg::ShowOk(GetWindow(),
                            L"Error while opening file!",
                            std::wstring(L"An error occured while opening file:\n")+
                            L"'"+fileName.GetPath()+L"'\n"+
                            L"Error:\n"+
                            L"'"+status.GetDescription()+L"'");

      return;
    }

    unsigned long tfid=htonl(trackingFileId);
    unsigned long tfv=htonl(trackingFileVersion);
    unsigned long tft=htonl(Lum::Base::SystemTime().GetTime());

    trackFile.write((const char*)&tfid,sizeof(tfid));
    trackFile.write((const char*)&tfv,sizeof(tfv));
    trackFile.write((const char*)&tft,sizeof(tft));

    startTrackingAction->Disable();
    stopTrackingAction->Enable();
    tracking->Set(true);
  }

  void AddTrackingRecord(const GPSData& data)
  {
    if (!tracking->Get()) {
      return;
    }

    double                lon=0.0;
    double                lat=0.0;
    double                alt=0.0;
    Lum::Base::SystemTime tim;

    if (data.HasLongitude()) {
      lon=data.GetLongitude();
    }

    if (data.HasLatitude()) {
      lat=data.GetLatitude();
    }

    if (data.HasAltitude()) {
      alt=data.GetAltitude();
    }

    if (data.HasTime()) {
      tim=data.GetTime();
    }

    unsigned long flon=htonl((unsigned long)rint((lon+90.0)*10000000.0));
    unsigned long flat=htonl((unsigned long)rint((lat+180.0)*10000000.0));
    unsigned long falt=htonl((unsigned long)rint(alt*100));
    unsigned long ftim=htonl(tim.GetTime());

    //std::cout << "TrackRecord: " << std::hex << flon << " " << flat << " " << falt << " " << ftim << std::endl;

    trackFile.write((const char*)&flon,sizeof(flon));
    trackFile.write((const char*)&flat,sizeof(flat));
    trackFile.write((const char*)&falt,sizeof(falt));
    trackFile.write((const char*)&ftim,sizeof(ftim));
  }

  void StopTracking()
  {
    trackFile.close();

    startTrackingAction->Enable();
    stopTrackingAction->Disable();
    tracking->Set(false);

    RefreshTracks();
  }

  void RefreshTracks()
  {
    Lum::Base::Path              dirName,fileName;
    Lum::Base::DirScannerRef     scanner;
    Lum::Base::DirScanner::Entry entry;
    Lum::Base::Status            status;

    tracks->Clear();
    tracks->Off();

    dirName.SetNativeDir(Lum::Base::Path::GetHome());
    dirName.AppendDir(L".gpsjinni");

    scanner=Lum::Base::DirScanner::Create(dirName);

    while (scanner->GetNext(entry,status)) {
      if (entry.isDir) {
        continue;
      }

      std::cout << "Found possible track '" << Lum::Base::WStringToString(entry.name) << "'" << std::endl;

      if (entry.name.length()<4 ||
          entry.name.substr(entry.name.length()-4)!=L".gjt") {
        std::cerr << "wrong suffix!" << std::endl;
        continue;
      }

      std::ifstream file;

      fileName=dirName;
      fileName.SetBaseName(entry.name);

      file.open(Lum::Base::WStringToString(fileName.GetPath()).c_str(),
                std::ios::in|std::ios::binary);

      if (!file) {
        std::cerr << "Cannot open file!" << std::endl;
        continue;
      }

      unsigned long tfid,tfv,tft;

      file.read((char*)&tfid,sizeof(tfid));
      file.read((char*)&tfv,sizeof(tfv));
      file.read((char*)&tft,sizeof(tft));

      if (!file) {
        std::cerr << "Cannot read file header!" << std::endl;
        continue;
      }

      tfid=ntohl(tfid);
      tfv=ntohl(tfv);
      tft=ntohl(tft);

      if (tfid!=trackingFileId) {
        std::cerr << "Wrong magic!" << std::endl;
        continue;
      }

      if (tfv!=trackingFileVersion) {
        std::cerr << "Wrong file version!" << std::endl;
        continue;
      }

      tracks->Append(Track(entry.name.substr(0,entry.name.length()-4),
                           Lum::Base::SystemTime(tft)));
    }

    tracks->Sort(1);

    tracks->On();
  }

  void ExportTrack(const std::wstring& name)
  {
    Lum::Base::Path                path;
    std::wstring                   file;
    Lum::Dlg::File::Options        *options=new Lum::Dlg::File::Options();
    Lum::Model::Dir::PatternFilter *filter;
    std::ifstream                  in;
    std::ofstream                  out;

    path.SetNativeDir(Lum::Base::Path::GetHome());
    path.AppendDir(L".gpsjinni");
    path.SetBaseName(name+L".gjt");

    in.open(Lum::Base::WStringToString(path.GetPath()).c_str(),
            std::ios::in|std::ios::binary);

    if (!in) {
      Lum::Base::Status status;

      status.SetToCurrentErrno();
      Lum::Dlg::Msg::ShowOk(GetWindow(),
                            L"Error while opening file!",
                            std::wstring(L"An error occured while opening file:\n")+
                            L"'"+file+L"'\n"+
                            L"Error:\n"+
                            L"'"+status.GetDescription()+L"'");

      return;
    }

    unsigned long tfid,tfv,tft;

    in.read((char*)&tfid,sizeof(tfid));
    in.read((char*)&tfv,sizeof(tfv));
    in.read((char*)&tft,sizeof(tft));

    if (!in) {
      std::cerr << "Cannot read file header!" << std::endl;
      return;
    }

    tfid=ntohl(tfid);
    tfv=ntohl(tfv);
    tft=ntohl(tft);

    if (tfid!=trackingFileId) {
      Lum::Dlg::Msg::ShowOk(GetWindow(),
                            L"Error while opening file!",
                            std::wstring(L"An error occured while opening file:\n")+
                            L"'"+file+L"'\n"+
                            L"Error:\n"+
                            L"'File is not a stored track'");
      return;
    }

    if (tfv!=trackingFileVersion) {
      Lum::Dlg::Msg::ShowOk(GetWindow(),
                            L"Error while opening file!",
                            std::wstring(L"An error occured while opening file:\n")+
                            L"'"+file+L"'\n"+
                            L"Error:\n"+
                            L"'File format version not supported'");
      return;
    }

    path.SetNativeDir(Lum::Base::Path::GetHome());
    path.SetBaseName(name+L".kml");

    file=path.GetPath();

    options->SetMode(Lum::Dlg::File::Options::modeFile);
    options->SetExistingOnly(false);

    filter=new Lum::Model::Dir::PatternFilter(L"KML Files");
    filter->AddPattern(L"*.kml");
    options->AddFilter(filter);

    if (!Lum::Dlg::File::SaveFile(GetWindow(),
                                  L"Exporting track...",
                                  file,
                                  options)) {
      return;
    }

    std::cout << "Saving to '" << Lum::Base::WStringToString(file) << "'" << std::endl;

    out.open(Lum::Base::WStringToString(file).c_str(),
             std::ios::out|std::ios::trunc);

    if (!out) {
      Lum::Base::Status status;

      status.SetToCurrentErrno();
      Lum::Dlg::Msg::ShowOk(GetWindow(),
                            L"Error while opening file!",
                            std::wstring(L"An error occured while opening file:\n")+
                            L"'"+file+L"'\n"+
                            L"Error:\n"+
                            L"'"+status.GetDescription()+L"'");

      return;
    }

    out << "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" << std::endl;
    out << "<kml xmlns=\"http://www.opengis.net/kml/2.2\">" << std::endl;
    out << "  <Document>" << std::endl;
    out << "    <name>" << Lum::Base::WStringToUTF8(name) << ".kml" << "</name>" << std::endl;
    out << "    <open>1</open>" << std::endl;
    out << "    <Style id=\"lineStyle\">" << std::endl;
    out << "      <LineStyle>" << std::endl;
    out << "        <color>7f0000ff</color>" << std::endl;
    out << "        <width>4</width>" << std::endl;
    out << "      </LineStyle>" << std::endl;
    out << "    </Style>" << std::endl;

    out << "    <Placemark>" << std::endl;
    out << "      <name>" << Lum::Base::WStringToUTF8(name) << "</name>" << std::endl;
    out << "      <styleUrl>#lineStyle</styleUrl>" << std::endl;
    out << "      <LineString>" << std::endl;
    out << "        <extrude>0</extrude>" << std::endl;
//    out << "        <tessellate>1</tessellate>" << std::endl;
    out << "        <altitudeMode>absolute</altitudeMode>" << std::endl;
    out << "        <coordinates>" << std::endl;

    while (in) {
      unsigned long flon;
      unsigned long flat;
      unsigned long falt;
      unsigned long ftim;

      in.read((char*)&flon,sizeof(flon));
      in.read((char*)&flat,sizeof(flat));
      in.read((char*)&falt,sizeof(falt));
      in.read((char*)&ftim,sizeof(ftim));

      if (in) {
        double                lon;
        double                lat;
        double                alt;
        Lum::Base::SystemTime time;

        flon=ntohl(flon);
        flat=ntohl(flat);
        falt=ntohl(falt);
        ftim=ntohl(ftim);

        lon=flon/10000000.0-90.0;
        lat=flat/10000000.0-180.0;
        alt=falt/100;
        time.SetTime(ftim);

        out << "          " << lon << "," << lat << "," << alt << std::endl;
      }
    }

    out << "        </coordinates>" << std::endl;
    out << "      </LineString>" << std::endl;
    out << "    </Placemark>" << std::endl;
    out << "  </Document>" << std::endl;
    out << "</kml>" << std::endl;

    out.close();
    in.close();
  }

  void DeleteTrack(const std::wstring& name)
  {
    Lum::Base::Path   path;
    Lum::Base::Status status;

    path.SetNativeDir(Lum::Base::Path::GetHome());
    path.AppendDir(L".gpsjinni");
    path.SetBaseName(name+L".gjt");

    status=Lum::Base::Path::RemoveFile(path.GetPath());

    if (status) {
      RefreshTracks();
    }
    else {
      Lum::Dlg::Msg::ShowOk(GetWindow(),
                            L"Error while deleting file!",
                            std::wstring(L"An error occured while deletion of file:\n")+
                            L"'"+path.GetPath()+L"'\n"+
                            L"Error:\n"+
                            L"'"+status.GetDescription()+L"'");
    }
  }

  void ShowChange(const GPSData& data)
  {
    if (data.HasConnected()) {
      connected->Set(data.GetConnected());
    }
    else {
      connected->Set(false);
    }

    if (data.HasOnline()) {
      online->Set(data.GetOnline());
    }
    else {
      online->Set(false);
    }

    if (data.HasFix()) {
      fix->Set(data.GetFix());
    }
    else {
      fix->Set(false);
    }

    if (data.HasTime()) {
      time->Set(data.GetTime().GetLocalLocaleDateTime());
    }
    else {
      time->SetNull();
    }

    if (data.HasLatitude()) {
      latitude->Set(DoubleToString(data.GetLatitude())+L"\u00b0");
    }
    else {
      latitude->SetNull();
    }

    if (data.HasLongitude()) {
      longitude->Set(DoubleToString(data.GetLongitude())+L"\u00b0");
    }
    else {
      longitude->SetNull();
    }

    if (data.HasAltitude()) {
      altitude->Set(DoubleToString(data.GetAltitude())+L"m");
    }
    else {
      altitude->SetNull();
    }

    if (data.HasSpeed()) {
      if (data.GetSpeed()>=1/3.6) {
        speed->Set(DoubleToString(data.GetSpeed()*3.6)+L"km/h");
      }
      else {
        speed->Set(DoubleToString(data.GetSpeed())+L"m/s");
      }
    }
    else {
      speed->SetNull();
    }

    if (data.HasTrack()) {
      track->Set(DoubleToString(data.GetTrack())+L"\u00b0");
      direction->Set(data.GetTrack());
    }
    else {
      track->SetNull();
      direction->SetNull();
    }

    if (data.HasClimb()) {
      climb->Set(DoubleToString(data.GetClimb())+L"m");
    }
    else {
      climb->SetNull();
    }

    if (data.HasSatellites()) {
      // Delete everything in model that is not in current list
      size_t i=1;
      while (i<=satellites->GetRows()) {
        bool found=false;

        for (size_t j=0; j<data.GetSatellites().size(); j++) {
          if (satellites->GetEntry(i).id==data.GetSatellites()[j].id) {
            found=true;
            break;
          }
        }

        if (found) {
          i++;
          continue;
        }
        else {
          satellites->Delete(i);
        }
      }

      // Now add everything to model that is not already in
      for (size_t i=0; i<data.GetSatellites().size(); i++) {
        bool found=false;

        for (size_t j=1; j<=satellites->GetRows(); j++) {
          if (satellites->GetEntry(j).id==data.GetSatellites()[i].id) {
            found=true;

            if (satellites->GetEntry(j).snr!=data.GetSatellites()[i].snr ||
                satellites->GetEntry(j).used!=data.GetSatellites()[i].used) {
              satellites->GetEntry(j)=data.GetSatellites()[i];
              satellites->RedrawRow(j);
            }
            break;
          }
        }

        if (!found) {
          satellites->Append(data.GetSatellites()[i]);
        }
      }
    }
    else {
      satellites->Clear();
    }
  }

  void UpdateStreams(const GPSData& data)
  {
    if (!data.HasFix() || !data.GetFix() ||
        !data.HasSpeed()) {
      speedStream->SetNull();
    }
    else {
      speedStream->Set(0,data.GetSpeed());
    }

    if (!data.HasFix() || !data.GetFix() ||
        !data.HasAltitude()) {
      altitudeStream->SetNull();
    }
    else {
      altitudeStream->Set(0,data.GetAltitude());
    }
  }

  void Resync(Lum::Base::Model* model, const Lum::Base::ResyncMsg& msg)
  {
    if (model==GetOpenedAction() && GetOpenedAction()->IsFinished()) {
      InitializeGPS();
      Lum::OS::display->AddTimer(refreshTime,0,refreshTimer);
      Lum::OS::display->AddTimer(reconnectTime,0,reconnectCheckTimer);
      RefreshTracks();
    }
    else if (model==GetClosedAction() && GetClosedAction()->IsFinished()) {
      DeinitializeGPS();
    }
    else if (model==aboutAction && aboutAction->IsFinished()) {
      Lum::Dlg::About::Show(this,info);
    }
    else if (model==dataChangedAction && dataChangedAction->IsFinished()) {
      GPSData dataCopy;

      GetGPSDataCopy(dataCopy);
      AddTrackingRecord(dataCopy);
      ShowChange(dataCopy);
    }
    else if (model==refreshTimer && refreshTimer->IsFinished()) {
      GPSData dataCopy;

      GetGPSDataCopy(dataCopy);
      ShowChange(dataCopy);
      AddTrackingRecord(dataCopy);
      UpdateStreams(dataCopy);

      Lum::OS::display->AddTimer(refreshTime,0,refreshTimer);
    }
    else if (model==reconnectCheckTimer && reconnectCheckTimer->IsFinished()) {
      ReconnectIfDisconnected();

      Lum::OS::display->AddTimer(reconnectTime,0,reconnectCheckTimer);
    }
    else if (model==startTrackingAction &&
             startTrackingAction->IsFinished() &&
             startTrackingAction->IsEnabled()) {
      StartTracking();
    }
    else if (model==stopTrackingAction &&
             stopTrackingAction->IsFinished() &&
             stopTrackingAction->IsEnabled()) {
      StopTracking();
    }
    else if (model==deleteTrackAction &&
             deleteTrackAction->IsFinished() &&
             deleteTrackAction->IsEnabled()) {
      DeleteTrack(tracks->GetEntry(trackSelection->GetLine()).GetName());
    }
    else if (model==exportTrackAction &&
             exportTrackAction->IsFinished() &&
             exportTrackAction->IsEnabled()) {
      ExportTrack(tracks->GetEntry(trackSelection->GetLine()).GetName());
    }
    else if (model==trackSelection) {
      if (trackSelection->HasSelection()) {
        deleteTrackAction->Enable();
        exportTrackAction->Enable();
      }
      else {
        deleteTrackAction->Disable();
        exportTrackAction->Disable();
      }
    }

    Dialog::Resync(model,msg);
  }

  void HandleChange()
  {
    Lum::OS::display->QueueActionForAsyncNotification(dataChangedAction);
  }
};

#if defined(HAVE_LIB_GPS)
static void cb(struct gps_data_t* data, char * buf, size_t len, int level)
{
  assert(window!=NULL);

  Lum::OS::WriteGuard<Lum::OS::RWMutex> guard(window->dataMutex);

  GPSData                               &gpsData=window->data;

  gpsData.SetConnected(true);

  if (data->set & ONLINE_SET) {
    gpsData.SetOnline(data->online);
  }

  if (data->set & STATUS_SET) {
    gpsData.SetFix(data->status!=STATUS_NO_FIX);
  }

  if ((data->set & LATLON_SET)  && !isnan(data->fix.latitude)) {
    gpsData.SetLatitude(data->fix.latitude);
  }

  if ((data->set & LATLON_SET)  && !isnan(data->fix.longitude)) {
    gpsData.SetLongitude(data->fix.longitude);
  }

  if ((data->set & ALTITUDE_SET) && !isnan(data->fix.altitude)) {
    gpsData.SetAltitude(data->fix.altitude);
  }

  if ((data->set & SPEED_SET) && !isnan(data->fix.speed)) {
    gpsData.SetSpeed(data->fix.speed/3.6);
  }

  if ((data->set & TRACK_SET) && !isnan(data->fix.track)) {
    gpsData.SetTrack(data->fix.track);
  }

  if ((data->set & CLIMB_SET) && !isnan(data->fix.climb)) {
    gpsData.SetClimb(data->fix.climb);
  }

  if (data->set & TIME_SET) {
    Lum::Base::SystemTime time(lround(data->fix.time));
    gpsData.SetTime(time);
  }

  if (data->set & SATELLITE_SET) {
    std::vector<GPSData::Satellite> satellites;

    for (int i=0; i<data->satellites; i++) {
      GPSData::Satellite s;

      s.id=data->PRN[i];
      s.used=data->used[i];
      s.snr=data->ss[i];

      satellites.push_back(s);
    }

    gpsData.SetSatellites(satellites);
  }

  window->HandleChange();
}
#elif defined(HAVE_LIB_LOCATION)
static void gps_location_changed(LocationGPSDevice *device, gpointer userdata)
{
  std::cout << "gps_location_changed!" << std::endl;
  assert(window!=NULL);

  Lum::OS::WriteGuard<Lum::OS::RWMutex> guard(window->dataMutex);

  GPSData                               &gpsData=window->data;

  gpsData.SetConnected(true);
  gpsData.SetOnline(device->online);
  gpsData.SetFix(device->status!=LOCATION_GPS_DEVICE_STATUS_NO_FIX);

  /* Fix information */
  LocationGPSDeviceFix *fix=device->fix;

  if (fix->fields & LOCATION_GPS_DEVICE_TIME_SET) {
    Lum::Base::SystemTime time(lround(fix->time));
    gpsData.SetTime(time);
  }
  if (fix->fields & LOCATION_GPS_DEVICE_LATLONG_SET && !isnan(fix->latitude)) {
    gpsData.SetLatitude(fix->latitude);
  }
  if (fix->fields & LOCATION_GPS_DEVICE_LATLONG_SET && !isnan(fix->longitude)) {
    gpsData.SetLongitude(fix->longitude);
  }
  if (fix->fields & LOCATION_GPS_DEVICE_ALTITUDE_SET && !isnan(fix->altitude)) {
    gpsData.SetAltitude(fix->altitude);
  }
  if (fix->fields & LOCATION_GPS_DEVICE_SPEED_SET && !isnan(fix->speed)) {
    gpsData.SetSpeed(fix->speed);
  }
  if (fix->fields & LOCATION_GPS_DEVICE_TRACK_SET && !isnan(fix->track)) {
    gpsData.SetTrack(fix->track);
  }
  if (fix->fields & LOCATION_GPS_DEVICE_CLIMB_SET && !isnan(fix->climb)) {
    gpsData.SetClimb(fix->climb);
  }

  std::vector<GPSData::Satellite> satellites;

  if (device->satellites!=NULL) {
    for (size_t i=0; i<device->satellites->len; i++) {
      LocationGPSDeviceSatellite *sat=(LocationGPSDeviceSatellite*)g_ptr_array_index(device->satellites,i);
      GPSData::Satellite         s;

      s.id=sat->prn;
      s.used=sat->in_use;
      s.snr=sat->signal_strength;

      satellites.push_back(s);
    }
  }

  gpsData.SetSatellites(satellites);

  window->HandleChange();
}

static void gps_location_connected(LocationGPSDevice *new_device, gpointer userdata) {
  std::cout << "Connected..." << std::endl;
}

static void gps_location_disconnected(LocationGPSDevice *new_device, gpointer userdata){
  std::cout << "disconnected..." << std::endl;
}

static void gps_location_started(LocationGPSDControl *control, gpointer userdata) {
  std::cout << "GPSD started!" << std::endl;
}

static void gps_location_stopped(LocationGPSDControl *control, gpointer userdata) {
  std::cout << "GPSD stoped!" << std::endl;
}

static void gps_location_error(LocationGPSDControl *control, gpointer userdata) {
  std::cout << "GPSD error!" << std::endl;
}
#endif

class Main : public Lum::OS::MainDialog<MyWindow>
{
public:
  Main()
  {
    info.SetProgram(Lum::Base::StringToWString(PACKAGE_NAME));
    info.SetVersion(Lum::Base::StringToWString(PACKAGE_VERSION));
    info.SetDescription(L"Where the #*~@# am I?");
    info.SetAuthor(L"Tim Teulings");
    info.SetContact(L"Tim Teulings <tim@teulings.org>");
    info.SetCopyright(L"(c) 2009, Tim Teulings");
    info.SetLicense(L"GNU Public License");
  }
};

LUM_MAIN(Main,Lum::Base::StringToWString(PACKAGE_NAME))

